-- $Id: TableExplorer.lua 16 2014-02-24 05:55:58Z diesal2010 $
local AddOnName, Env = ...
local ADDON = {}
Env[1], _G[AddOnName] = ADDON, ADDON
-- ~~| Libraries |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local DiesalTools 	= LibStub("DiesalTools-1.0")
local DiesalStyle 	= LibStub("DiesalStyle-1.0") 
local DiesalGUI 		= LibStub("DiesalGUI-1.0")
-- ~~| Lua Upvalues |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local print, type, select, tostring, tonumber		= print, type, select, tostring, tonumber
local sub, format, match, lower							= string.sub, string.format, string.match, string.lower
local tsort														= table.sort
local abs														= math.abs
-- ~~| WoW Upvalues |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

-- ~~| Slash Command |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
SLASH_TEXPLORE1= '/texplore'
function SlashCmdList.TEXPLORE(msg) 
 local table, depth = msg:match("^([^%s]*)%s*(%d*).*$")	
	if not _G[table] then print('table dosnt exist') return end 	
	texplore(table,_G[table],tonumber(depth))	
end
-- ~~| Explorer Widget Pool |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local tableExplorerPool = {}
local function newTableExplorer()	
	
	local newTableExplorer = next(tableExplorerPool)
	if not newTableExplorer then
		newTableExplorer = ADDON:ConstructTableExplorer()		
	else
		tableExplorerPool[newTableExplorer] = nil		
	end
	
	newTableExplorer:OnAcquire()
	
	return newTableExplorer	
end
local function releaseTableExplorer(tableExplorer)
	tableExplorer:OnRelease()	
	tableExplorerPool[tableExplorer] = true	
end
-- ~~| Global Explorer Call |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function texplore(tname,t,depth)
	local tableExplorer = newTableExplorer()
	tableExplorer:SetTable(tname,t,depth)	
end
-- ~~| TableExplorer StyleSheets |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local expandButtonStyle = {
	type			= 'texture',
	texFile		= 'DiesalGUIcons',
	texCoord		= {1,6,16,256,128},
	alpha 		= .7,	
	offset		= {-2,nil,-2,nil},
	width			= 16,
	height		= 16,			
}
local expandButtonNormal = {
	type			= 'texture',
	texColor		= 'ffffff',	
	alpha 		= .7,			
}
local expandButtonOver = {
	type			= 'texture',	
	alpha 		= 1,		
} 
local expandButtonDisabled = {
	type			= 'texture',	
	alpha 		= .3,
}
local collapseButtonStyle = {
	type			= 'texture',
	texFile		='DiesalGUIcons',
	texCoord		= {2,6,16,256,128},
	alpha 		= .7,	
	offset		= {-2,nil,-2,nil},
	width			= 16,
	height		= 16,			
}
local collapseButtonNormal = {
	type			= 'texture',	
	alpha 		= .7,			
}
local collapseButtonOver = {
	type			= 'texture',	
	alpha 		= 1,		
} 
local collapseButtonDisabled = {
	type			= 'texture',	
	alpha 		= .3,
}
local refreshButtonStyle = {			
	type			= 'texture',
	texFile		='DiesalGUIcons',
	texCoord		= {3,5,16,256,128},
	alpha 		= .7,	
	offset		= {-2,nil,-2,nil},
	width			= 16,
	height		= 16,			
}
local refreshButtonNormal = {
	type			= 'texture',	
	alpha 		= .7,			
}
local refreshButtonOver = {
	type			= 'texture',	
	alpha 		= 1,		
} 
local refreshButtonDisabled = {
	type			= 'texture',	
	alpha 		= .3,
}
local homeButtonStyle = {
	type			= 'texture',
	texFile		='DiesalGUIcons',
	texCoord		= {4,5,16,256,128},
	alpha 		= .7,	
	offset		= {-2,nil,-2,nil},
	width			= 16,
	height		= 16,			
}
local homeButtonNormal = {
	type			= 'texture',	
	alpha 		= .7,			
}
local homeButtonOver = {
	type			= 'texture',	
	alpha 		= 1,		
} 
local homeButtonDisabled = {
	type			= 'texture',	
	alpha 		= .3,
}
-- ~~| TableExplorer Locals |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local TIMEOUT = 5
-- ~~| TableExplorer Methods |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local methods = {	
	['OnAcquire'] = function(self)				
		self.settings = DiesalTools:TableCopy( self.defaults )			
		self:ApplySettings()				
		self:Show()		
	end,
	['OnRelease'] = function(self)				
		self.scrollFrame:SetContentHeight(10)
		self.tree:ReleaseChildren()
	end,		
	['SetSettings'] = function(self,settings,apply)
		for key,value in pairs(settings) do
			self.settings[key] = value
		end
		if apply then self:ApplySettings()	end
	end,
	['ApplySettings'] = function(self)
		
	end,	
	['SetTable'] = function(self,tname,t,depth)
		if type(t) ~='table' then return end
		local depth = depth or self.tree.defaults.exploreDepth	
		
		self.depthSpinner:SetNumber(depth)
		self.timeSpinner:SetNumber(TIMEOUT) -- Default timeout
		
		self.window:SetTitle('|cffffaa00Table Explorer',tname or '')				
			
		-- self:RefreshTable()
		self.tree:SetSettings({
			timeout						= TIMEOUT,
			exploreDepth 				= depth,
			homeTable 					= t,
			homeTableName				= tname,					
		})		
		self.tree:BuildTree(tname,t)		
	end,
	['Show'] = function(self)
		self.window:Show()
	end,		
	['AddHistory'] = function(self,t,tname)
		self.history = self.settings.history
		if type(t) ~='table' then return end			
		local s = tostring(t)
		self.history[#self.history+1].t = t
		self.history[#self.history+1].t = tname or s	
	end,						
}
-- ~~| TableExplorer Constructor |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function ADDON:ConstructTableExplorer()	
	local self = {}	
	-- ~~ Default Settings ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	self.defaults = {		
			
	}
	-- ~~ Events ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	-- OnAcquire, OnRelease, OnHeightSet, OnWidthSet	
	-- ~~ Construct ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	local window = DiesalGUI:Create('Window')	
	window:SetSettings({
		header			= true,
		footer			= true,
		top				= 600,
		left 				= 5,
		headerHeight	= 21,
		height 			= 400,
		width 			= 300,
		minWidth 		= 250,		
		sizerBRHeight	= 20,
		sizerBRWidth	= 20,		
	},true)	
	window:SetEventListener('OnHide', function(this,event) releaseTableExplorer(self) end)		
	
	local homeButton = DiesalGUI:Create('Button')
	homeButton:SetParent(window.header)
	homeButton:SetPoint('TOPLEFT',0,0)	
	homeButton:SetSettings({		
		width			= 20,		
		height		= 20,
	},true)
	homeButton:SetStyle('frame',homeButtonStyle)	
	homeButton:SetEventListener('OnEnter', 	function()
		homeButton:SetStyle('frame',homeButtonOver)
		
		GameTooltip:SetOwner(homeButton.frame, "ANCHOR_TOPLEFT",0,2) 		
		GameTooltip:AddLine('Base Table')		
		GameTooltip:Show()
	end)		
	homeButton:SetEventListener('OnLeave', 	function()
		homeButton:SetStyle('frame',homeButtonNormal)
		GameTooltip:Hide()
	end)		
	homeButton:SetEventListener('OnDisable', 	function() homeButton:SetStyle('frame',homeButtonDisabled)	end)		
	homeButton:SetEventListener('OnEnable', 	function() homeButton:SetStyle('frame',homeButtonNormal)		end)
	homeButton:SetEventListener('OnClick', 	function() 
		local settings = self.tree.settings		
		self.tree:BuildTree(settings.homeTableName,settings.homeTable)
	end)
	window:AddChild(homeButton)
	
	local refreshButton = DiesalGUI:Create('Button')
	refreshButton:SetParent(window.header)
	refreshButton:SetPoint('TOPLEFT',20,0)	
	refreshButton:SetSettings({		
		width			= 20,		
		height		= 20,
	},true)
	refreshButton:SetStyle('frame',refreshButtonStyle)	
	refreshButton:SetEventListener('OnEnter', 	function()
		refreshButton:SetStyle('frame',refreshButtonOver)
		
		GameTooltip:SetOwner(refreshButton.frame, "ANCHOR_TOPLEFT",0,2) 		
		GameTooltip:AddLine('Refresh Table')		
		GameTooltip:Show()
	end)		
	refreshButton:SetEventListener('OnLeave', 	function()
		refreshButton:SetStyle('frame',refreshButtonNormal)
		GameTooltip:Hide()
	end)	
	refreshButton:SetEventListener('OnDisable', 	function() refreshButton:SetStyle('frame',refreshButtonDisabled)	end)		
	refreshButton:SetEventListener('OnEnable', 	function() refreshButton:SetStyle('frame',refreshButtonNormal)		end)
	refreshButton:SetEventListener('OnClick', 	function() self.tree:BuildTree() end)	
	window:AddChild(refreshButton)	
	
	local collapseButton = DiesalGUI:Create('Button')
	collapseButton:SetParent(window.header)
	collapseButton:SetPoint('TOPLEFT',40,0)	
	collapseButton:SetSettings({		
		width			= 20,		
		height		= 20,
	},true)
	collapseButton:SetStyle('frame',collapseButtonStyle)	
	collapseButton:SetEventListener('OnEnter', 	function()
		collapseButton:SetStyle('frame',collapseButtonOver)
		
		GameTooltip:SetOwner(collapseButton.frame, "ANCHOR_TOPLEFT",0,2) 		
		GameTooltip:AddLine('Collapse All')		
		GameTooltip:Show()
	end)		
	collapseButton:SetEventListener('OnLeave', 	function()
		collapseButton:SetStyle('frame',collapseButtonNormal)
		GameTooltip:Hide()
	end)	
	collapseButton:SetEventListener('OnDisable', 	function() collapseButton:SetStyle('frame',collapseButtonDisabled)	end)		
	collapseButton:SetEventListener('OnEnable', 	function() collapseButton:SetStyle('frame',collapseButtonNormal)			end)	
	collapseButton:SetEventListener('OnClick', 	function() self.tree:CollapseAll() end)		
	window:AddChild(collapseButton)
	
	local expandButton = DiesalGUI:Create('Button')
	expandButton:SetParent(window.header)
	expandButton:SetPoint('TOPLEFT',60,0)	
	expandButton:SetSettings({		
		width			= 20,		
		height		= 20,
	},true)
	expandButton:SetStyle('frame',expandButtonStyle)	
	expandButton:SetEventListener('OnEnter', 	function()
		expandButton:SetStyle('frame',expandButtonOver)
		
		GameTooltip:SetOwner(expandButton.frame, "ANCHOR_TOPLEFT",0,2) 		
		GameTooltip:AddLine('Expand All')		
		GameTooltip:Show()
	end)		
	expandButton:SetEventListener('OnLeave', 	function()
		expandButton:SetStyle('frame',expandButtonNormal)
		GameTooltip:Hide()
	end)	
	expandButton:SetEventListener('OnDisable', 	function() expandButton:SetStyle('frame',expandButtonDisabled)	end)		
	expandButton:SetEventListener('OnEnable', 	function() expandButton:SetStyle('frame',expandButtonNormal)	end)
	expandButton:SetEventListener('OnClick', 	function() self.tree:ExpandAll() end)		
	window:AddChild(expandButton)
	
	local depthText = window.header:CreateFontString(nil,"OVERLAY",'DiesalFontNormal')	 	
	depthText:SetPoint('TOPRIGHT',-113,-1)	
	depthText:SetSize(100,16)
	depthText:SetJustifyH("RIGHT")
	depthText:SetJustifyV("BOTTOM")
	depthText:SetWordWrap(false)
	depthText:SetTextColor(1,1,1,.7)
	depthText:SetText('Depth:')	 	
	
	local depthSpinner = DiesalGUI:Create('Spinner')
	depthSpinner:SetParent(window.header)
	depthSpinner:SetPoint('TOPRIGHT',-82,-2)	
	depthSpinner:SetSettings({
		min	= 1,		
		max 	= 10,		
	})	
	depthSpinner:SetEventListener('OnValueChanged', function(this,event,userInput,number) self.tree:SetSettings({exploreDepth = number})	end)
	depthSpinner:SetEventListener('OnEnter', 	function()		
		
		GameTooltip:SetOwner(depthSpinner.frame, "ANCHOR_TOPLEFT",0,2) 		
		GameTooltip:AddLine('Set Table Exploration Depth')		
		GameTooltip:Show()
	end)		
	depthSpinner:SetEventListener('OnLeave', 	function()	GameTooltip:Hide() end)	
	window:AddChild(depthSpinner)	
	
	local timeText = window.header:CreateFontString(nil,"OVERLAY",'DiesalFontNormal')	 	
	timeText:SetPoint('TOPRIGHT',-33,-1)	
	timeText:SetSize(100,16)
	timeText:SetJustifyH("RIGHT")
	timeText:SetJustifyV("BOTTOM")
	timeText:SetWordWrap(false)
	timeText:SetTextColor(1,1,1,.7)
	timeText:SetText('Timeout:')	 	
	
	local timeSpinner = DiesalGUI:Create('Spinner')
	timeSpinner:SetParent(window.header)
	timeSpinner:SetPoint('TOPRIGHT',-2,-2)	
	timeSpinner:SetSettings({
		min	= 2,		
		max 	= 10,		
	})	
	timeSpinner:SetEventListener('OnValueChanged', function(this,event,userInput,number) self.tree:SetSettings({timeout = number})	end)
	timeSpinner:SetEventListener('OnEnter', 	function()		
		
		GameTooltip:SetOwner(timeSpinner.frame, "ANCHOR_TOPLEFT",0,2) 		
		GameTooltip:AddLine('Set Table Exploration Timeout (seconds)')		
		GameTooltip:Show()
	end)		
	timeSpinner:SetEventListener('OnLeave', 	function()	GameTooltip:Hide() end)	
	window:AddChild(timeSpinner)		
	
	local scrollFrame = DiesalGUI:Create('ScrollFrame')
	scrollFrame:SetParentObject(window)
	scrollFrame:SetSettings({contentPadLeft = 1},true)
	scrollFrame:SetPoint('TOPLEFT',0,-2)
	scrollFrame:SetPoint('BOTTOMRIGHT',-3,2)
	window:AddChild(scrollFrame)
	
	local tree = DiesalGUI:Create('TableExplorerTree')
	tree:SetParentObject(scrollFrame)	
	tree:SetEventListener('OnHeightChange', function(this,event,height) scrollFrame:SetContentHeight(height) end)
	tree:SetEventListener('OnStatusChange', function(this,event,text) self.statusText:SetText(text) end)	
	tree:SetAllPoints()
	scrollFrame:AddChild(tree)
	
	local statusText = window.footer:CreateFontString(nil,"OVERLAY",'DiesalFontNormal')	 	
	statusText:SetPoint('TOPLEFT',4,-7)	
	statusText:SetPoint('BOTTOMRIGHT',-6,0)	
	statusText:SetJustifyH("LEFT")
	statusText:SetJustifyV("TOP")
	statusText:SetWordWrap(false)
	statusText:SetTextColor(1,1,1,.7)
	
	-- ~~ Frames ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	self.window 			= window
	self.scrollFrame 		= scrollFrame	
	
	self.expandButton 	= expandButton
	self.collapseButton 	= collapseButton
	self.refreshButton 	= refreshButton
	self.homeButton 		= homeButton
	
	self.depthSpinner 	= depthSpinner
	self.timeSpinner 		= timeSpinner
	
	self.tree 				= tree
	
	self.statusText 		= statusText
	-- ~~ Methods ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	for method, func in pairs(methods) do
		self[method] = func
	end	
	-- ~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~		
	return self
end