-- $Id: tableExplorerBranch.lua 16 2014-02-24 05:55:58Z diesal2010 $
local AddOnName, Env = ... local ADDON = Env[1]
-- ~~| Libraries |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local DiesalTools = LibStub('DiesalTools-1.0')
local DiesalStyle = LibStub('DiesalStyle-1.0')
local DiesalGUI 	= LibStub('DiesalGUI-1.0')
local DiesalMenu 	= LibStub('DiesalMenu-1.0')
-- ~~| Lua Upvalues |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local sub, format, match, lower							= string.sub, string.format, string.match, string.lower
local tsort														= table.sort
local tostring													= tostring
-- ~~| WoW Upvalues |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

-- ~~| TableExplorerBranch |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local Type 		= 'TableExplorerBranch'
local Version 	= 1
-- ~~| TableExplorerBranch StyleSheets |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local styleSheet = {
	['button-highlight'] = {			
		type			= 'texture',
		layer			= 'HIGHLIGHT',						
		color			= 'ffffff',			
		alpha			= .1,		
	},	
}
local wireFrameSheet = {
	['frame-white'] = {	
		type			= 'outline',
		layer			= 'OVERLAY',						
		color			= 'ffffff',	
		alpha			= .15,			
	},
	['button-purple'] = {		
		type			= 'outline',
		layer			= 'OVERLAY',						
		color			= 'aa00ff',	
		alpha			= 0,		
	},
	['content-yellow'] = {				
		type			= 'outline',
		layer			= 'OVERLAY',						
		color			= 'fffc00',
		alpha			= 0,				
	},	
}

-- ~~| TableExplorerBranch Locals |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local BLUE			= DiesalTools:GetTxtColor('00aaff')
local DARKBLUE		= DiesalTools:GetTxtColor('004466')
local ORANGE		= DiesalTools:GetTxtColor('ffaa00')
local DARKORANGE	= DiesalTools:GetTxtColor('4c3300')
local GREY			= DiesalTools:GetTxtColor('7f7f7f')
local DARKGREY		= DiesalTools:GetTxtColor('414141')
local WHITE			= DiesalTools:GetTxtColor('e6e6e6')
local RED			= DiesalTools:GetTxtColor('ff0000')
local GREEN			= DiesalTools:GetTxtColor('00ff2b')
local YELLOW		= DiesalTools:GetTxtColor('ffff00')
local LIGHTYELLOW	= DiesalTools:GetTxtColor('ffea7f')

local function compare(a,b)
	if a[1] == b[1] then
		return a[2] < b[2]
	else
		return  a[1] <  b[1]
	end
end
local function sortType(value)		
	if type(value) == 'table' then return 1 end
	if type(value) == 'function' then return 2 end
	return 3
end
local function sortTable(t)
	local sortedTable = {}			
	for key, value in pairs(t) do			
		sortedTable[#sortedTable + 1] = {sortType(value),lower(tostring(key)), key  }							
	end		
	tsort(sortedTable, compare)
	return sortedTable		
end
-- ~~| TableExplorerBranch Methods |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local methods = {
	['OnAcquire'] = function(self)		
		self:ApplySettings()		
		self:AddStyleSheet(styleSheet)
		self:Show()
	end,
	['OnRelease'] = function(self)
		
	end,
	['ApplySettings'] = function(self)
		if not self.settings.key then return false end					
		
		local button			= self.button
		local content			= self.content				
		
		self:SetText()		
		self:ClearAllPoints()
		self[self.settings.expanded and "Expand" or "Collapse"](self)	
		button:SetHeight(self.settings.buttonHeight)		
				 
		if self.settings.position == 1 then
			self:SetPoint('TOPLEFT',self.settings.indent,0) 
			self:SetPoint('RIGHT')			
		else	
			local anchor = self.settings.parentObject.children[self.settings.position-1].frame
			self:SetPoint('TOPLEFT',anchor,'BOTTOMLEFT',0,0)
			self:SetPoint('RIGHT')			
		end			
	end,
	['BuildBranch'] = function(self,tree,key,value,position,depth)
		local settings 	= self.settings
		-- Set Settings
		settings.tree				= tree
		settings.exploreDepth	= tree.settings.exploreDepth
		settings.key				= key
		settings.value				= value
		settings.position			= position
		settings.depth				= depth
		-- Build Branch		
		self:ReleaseChildren()
		if value == tree or depth >= settings.exploreDepth or type(value) ~= 'table' then self:ApplySettings() return end
		local sortedTable = sortTable(value)
		for position, k in ipairs(sortedTable) do
		if tree.settings.endtime <= time() then										
				tree:ReleaseChildren()
				tree:UpdateHeight()
				tree:FireEvent("OnStatusChange",'|cffff0000Table Exploration Timed out.') 				
		return end			
			local key = k[3] 				
			local branch = DiesalGUI:Create('TableExplorerBranch')
			branch:SetParentObject(self)
			self:AddChild(branch)
			if position == #sortedTable then branch:SetSettings({last=true}) end				
			branch:BuildBranch(tree,key,value[key],position,depth + 1)
		end
		self:ApplySettings()
	end,	
	['Collapse'] = function(self)		
		self.settings.expanded = false
		self:DrawIcon()
		self:DrawFolds()				
		self.content:Hide()
		self:UpdateHeight()		
	end,
	['Expand'] = function(self)
		self.settings.expanded = true
		self:DrawIcon()
		self:DrawFolds()				
		self.content:Show()		
		self:UpdateHeight()
	end,	
	['UpdateHeight'] = function(self)
		local contentHeight = 0
		local settings = self.settings
		local children = self.children
		
		if settings.expanded then 
			for i=1 , #children do 
				contentHeight = contentHeight + children[i].frame:GetHeight()				
			end			
		end		
		self.content:SetHeight(contentHeight)		
		self:SetHeight(settings.buttonHeight + contentHeight)		
		settings.parentObject:UpdateHeight()
	end,	
	['FormatValue'] = function(self,value,type)
		if not value 			then return  format("%s%s",BLUE,'nil')		end
		if type == 'string' 	then return format("%s'%s'",GREEN,DiesalTools:EscapeString(value))	end
		if type == 'number' 	then return format("%s%s",YELLOW,value)				end
		if type == 'boolean' then return format("%s%s",BLUE,tostring(value))		end		
		
		local id = match(tostring(value),'.*: (.*)')
		id = id or tostring(value)
		
		if type == 'function' 	then return format('%sfunction%s %s',BLUE,DARKBLUE,id) 	end		
		if type == 'userdata' 	then return format('%suserdata%s %s',GREY,DARKGREY,id)	end
		
		if type == 'table' 		then
			if next(value) == nil then return format('%s{ } table%s %s',ORANGE,DARKORANGE,id) end -- Blank Table
			if self.settings.depth >= self.settings.exploreDepth then return format('%s{%s ... %s} table%s %s',ORANGE,RED,ORANGE,DARKORANGE,id) end -- explore depth reached 
			return format('%stable%s %s',ORANGE,DARKORANGE,id)		
		end						
	end,
	['FormatKey'] = function(self,key,type,valueType)
		if valueType == 'userdata' then return ''	end		
		
		if type == 'string' then
			if valueType == 'function' then
			 	return format('%s%s',BLUE,DiesalTools:EscapeString(key))				
			end
			if valueType == 'table' then				
				return format('%s%s',ORANGE,DiesalTools:EscapeString(key))
			end
			return format('%s%s',WHITE,DiesalTools:EscapeString(key))						
		end		
		
		if type == 'number' 	then return format("%s[%s%s%s]",ORANGE,YELLOW,key,ORANGE)	end
		if type == 'boolean' then return format("%s%s",BLUE,tostring(key))	end		
		
		local id = match(tostring(key),'.*: (.*)')
		id = id or tostring(key) 
		
		if type == 'function' 	then return format('%sfunction%s %s',BLUE,DARKBLUE,id) 	end		
		if type == 'userdata' 	then return format('%suserdata%s %s',GREY,DARKGREY,id)	end
		if type == 'table' 		then return format('%stable%s %s',ORANGE,DARKORANGE,id)	end	
	end,	
	['SetText'] = function(self)
		local settings		= self.settings		
		local valueType 	= type(settings.value)	
		local keyType		= type(settings.key)			
		
		local keyText 		= self:FormatKey(settings.key,keyType,valueType)		
		local valueText 	= self:FormatValue(settings.value,valueType)		
		self.text:SetText(format('%s%s = %s',keyText,LIGHTYELLOW,valueText))				
	end,
	['DrawIcon'] = function(self)
		local settings		= self.settings		
		local valueType 	= type(settings.value)
		
		if valueType == 'string' 	then self.typeIcon:SetTexCoord(DiesalTools:GetIconCoords(3,8,16,256,128)) return end
		if valueType == 'boolean' 	then self.typeIcon:SetTexCoord(DiesalTools:GetIconCoords(3,8,16,256,128)) return end
		if valueType == 'number' 	then self.typeIcon:SetTexCoord(DiesalTools:GetIconCoords(3,8,16,256,128)) return end
		if valueType == 'function' then self.typeIcon:SetTexCoord(DiesalTools:GetIconCoords(4,8,16,256,128)) return end	
		
		if valueType == 'table' then
			if next(settings.value) == nil or settings.depth >= settings.exploreDepth then								
				self.typeIcon:SetTexCoord(DiesalTools:GetIconCoords(1,8,16,256,128))
			return end
			
			if settings.expanded then 
				self.typeIcon:SetTexCoord(DiesalTools:GetIconCoords(2,8,16,256,128))
			else
				self.typeIcon:SetTexCoord(DiesalTools:GetIconCoords(1,8,16,256,128))
			end			
		return end		
		self.typeIcon:SetTexCoord(DiesalTools:GetIconCoords(3,8,16,256,128))	
										
	end,
	['DrawFolds'] = function(self)		
		local settings 	= self.settings
		local parent 		= self.parentObject
		local valueType	= type(settings.value)		
		local foldColor	= '353535'
		local expandColor	= '5a5a5a'
		
		-- Frame Fold Line
		if not settings.last and valueType == 'table' and next(settings.value) ~= nil and settings.depth < settings.exploreDepth then			
			self:SetStyle('frame-foldLine',{
				type			= 'texture',
				layer			= 'OVERLAY',						
				color			= foldColor,
				offset		= {6,nil,-11,2},
				width			= 1,				
			})	
		end
		-- Table Fold Lines
		if valueType == 'table' and next(settings.value) ~= nil and settings.depth < settings.exploreDepth then 
			self:SetStyle('button-square',{
				type			= 'outline',
				layer			= 'BORDER',						
				color			= foldColor,
				offset		= {10,nil,-2,nil},
				width			= 9,
				height		= 9,				
			})	
			self:SetStyle('button-expandH',{
				type			= 'texture',
				layer			= 'BORDER',						
				color			= expandColor,
				offset		= {8,nil,-6,nil},
				width			= 5,
				height		= 1,							
			})
			self:SetStyle('button-expandV',{
					type			= 'texture',
					layer			= 'BORDER',						
					color			= expandColor,
					offset		= {6,nil,-4,nil},
					width			= 1,
					height		= 5,
					alpha			= 1,							
			})		
			if settings.expanded then 
				self:SetStyle('button-expandV',{
					type			= 'texture',
					alpha			= 0,	
				})	
			end			
		else -- Non and Empty Table Lines
			self:SetStyle('button-lineV',{
				type			= 'texture',
				layer			= 'BORDER',						
				color			= foldColor,
				offset		= {6,nil,-6,nil},
				height		= 1,
				width			= 6,				
			})	
			self:SetStyle('button-lineH',{
				type			= 'texture',
				layer			= 'BORDER',						
				color			= foldColor,
				offset		= {6,nil,0,0},
				width			= 1,				
			})	
			if settings.last then			
				self:SetStyle('button-lineH',{
					type			= 'texture',					
					offset		= {6,nil,0,-7},								
				})					
			end	
		end
		
	end,
	['OnClick'] = function(self,button)
		if button =='LeftButton' then	
			self[not self.settings.expanded and "Expand" or "Collapse"](self)								
		elseif button =='RightButton' then			
			local menuData = self:BuildItemMenu()
			if not next(menuData) then return end
			DiesalMenu:Menu(menuData,self.frame,10,-10)				
		end				
	end,
	['BuildItemMenu'] = function(self)
		local menuData = {}
		local settings 		= self.settings
		local tree 				= settings.tree
		local children 		= self.children		
		local value				= settings.value
		local key				= settings.key		
		--[[ Menu Structure
			Explore Value			
			Explore Value in New Window
			Explore key			
			Explore Key in New Window
			Run Function
			---------------------------
			Collapse Children 			
		]]
		-- Explore Value 
		if type(value) == 'table' and next(value) then
			menuData[#menuData+1] = { name = 'Explore Table',				
				onClick 	= function() tree:BuildTree(type(key) == 'string' and tostring(key) or type(key) == 'number' and tostring(key) or tostring(value),value) end,			
			}
			menuData[#menuData+1] = { name = 'Explore Table in New Window',							
				onClick 	= function() texplore(type(key) == 'string' and tostring(key) or type(key) == 'number' and tostring(key) or tostring(value),value) end,
			}				
		end
		-- Explore key	
		if type(key) == 'table' and next(key) then
			menuData[#menuData+1] = { name = 'Explore Table',
				onClick 	= function() tree:BuildTree(tostring(key),key) end,					
			}
			menuData[#menuData+1] = { name = 'Explore Table in New Window',
				onClick 	= function() texplore(tostring(key),key) end,
			}				
		end
		-- Collapse Children 
		if type(value) == 'table' and next(value) and settings.depth < (settings.exploreDepth-1) then
			menuData[#menuData+1] = { type = 'spacer'}			
			menuData[#menuData+1] = { name = 'Collapse Children',
				onClick 	= function()					
					for i=1 , #children do 
						children[i]:Collapse()																	
					end					
				end,
			}
		end
		-- function 
		if type(value) == 'function' then					
			menuData[#menuData+1] = { name = 'Run Function',
				onClick 	= function() value()	end,
			}
		end
			
		return menuData
	end	
}
-- ~~| TableExplorerBranch Constructor |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local function Constructor()
	local self 		= DiesalGUI:GetObjectBase(Type)
	local frame 	= self.frame	
	-- ~~ Default Settings ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	self.defaults = {
		fontObject		= DiesalFontNormal,
		expanded			= true,			
		buttonHeight 	= 14,
		indent			= 13,			
	}
	-- ~~ Events ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	-- OnAcquire, OnRelease, OnHeightSet, OnWidthSet	
	-- ~~ Construct ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	
	local button = self:CreateRegion("Button", 'button', frame)			
	button:SetPoint('TOPRIGHT')
	button:SetPoint('TOPLEFT')
	button:RegisterForClicks('RightButtonUp','LeftButtonUp')			
	button:SetScript("OnClick", function(this,button) DiesalGUI:OnMouse(this,button); self:OnClick(button); end)	
	
	local typeIcon = self:CreateRegion("Texture", 'typeIcon', button)
	DiesalStyle:StyleTexture(typeIcon,{
		offset 	= {0,nil,2,nil},
		height	= 16,
		width		= 16,
		texFile	= 'DiesalGUIcons',			
	})		
	
	local text = self:CreateRegion("FontString", 'text', button)	
	text:SetPoint("TOPLEFT",14,-1) 
	text:SetPoint("BOTTOMRIGHT",-4,0)	
	text:SetHeight(0)
	text:SetJustifyH("TOP")
	text:SetJustifyH("LEFT")
	text:SetWordWrap(false)			
	
	local content = self:CreateRegion("Frame", 'content', button)		
	content:SetPoint('TOPLEFT',button,'BOTTOMLEFT',0,0)
	content:SetPoint('TOPRIGHT')
	content:SetHeight(0)	
	-- ~~ Methods ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	for method, func in pairs(methods) do
		self[method] = func
	end
	-- ~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	DiesalGUI:RegisterObject(self)
	return self
end
DiesalGUI:RegisterObjectConstructor(Type,Constructor,Version)
